"""Module contain components for work with db table."""
from __future__ import annotations

from numbers import Number
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Sequence

from ..sql import func


if TYPE_CHECKING:
    from lildb.operations import Query

    from ..sql import SQLBase
    from ..sql import TFunc
    from .table import Table


__all__ = (
    "ResultComparison",
    "Column",
    "Columns",
)


class ResultComparison(str):
    """The result of the comparison."""

    def _create_operation(
        self,
        operation: str,
        value: ResultComparison | str | SQLBase,
    ) -> ResultComparison:
        return ResultComparison(f"{value} {operation} {self}")

    def __and__(
        self,
        value: ResultComparison | str | SQLBase,
    ) -> ResultComparison:
        """."""
        return self._create_operation("AND", value)

    def __or__(
        self,
        value: ResultComparison | str | SQLBase,
    ) -> ResultComparison:
        """."""
        return self._create_operation("OR", value)

    def __rand__(
        self,
        value: ResultComparison | str | SQLBase,
    ) -> ResultComparison:
        """."""
        return self._create_operation("AND", value)

    def __ror__(
        self,
        value: ResultComparison | str | SQLBase,
    ) -> ResultComparison:
        """."""
        return self._create_operation("OR", value)


class Column:
    """Column."""

    __slots__ = (
        "table_name",
        "_column_name",
        "_full_column_name",
        "complete_label",
    )

    def __init__(self, table_name: str, column_name: str) -> None:
        """Initialize."""
        self.table_name = table_name
        self._column_name = column_name
        self.complete_label = column_name
        self._full_column_name = "`{}`.{}".format(
            self.table_name,
            self._column_name,
        )

    def _create_operation(
        self,
        operation: str,
        value: Any,
        *,
        in_operation: bool = False
    ) -> ResultComparison:
        """."""
        if in_operation:
            result = "{} {} {}".format(
                self._full_column_name,
                operation,
                value,
            )
            return ResultComparison(result)
        if isinstance(value, str):
            result = "{} {} '{}'".format(
                self._full_column_name,
                operation,
                value,
            )
            return ResultComparison(result)
        if value is None and operation == "=":
            return ResultComparison(
                "{} is NULL".format(
                    self._full_column_name,
                )
            )
        elif value is None:
            value = "NULL"
        result = "{} {} {}".format(
            self._full_column_name,
            operation,
            value,
        )
        return ResultComparison(result)

    def __eq__(self, value: object) -> ResultComparison:  # type: ignore
        """."""
        return self._create_operation("=", value)

    def __ne__(self, value: object) -> ResultComparison:  # type: ignore
        """."""
        return self._create_operation("!=", value)

    def __lt__(self, value: object) -> ResultComparison:
        """."""
        return self._create_operation("<", value)

    def __le__(self, value: object) -> ResultComparison:
        """."""
        return self._create_operation("<=", value)

    def __gt__(self, value: object) -> ResultComparison:
        """."""
        return self._create_operation(">", value)

    def __ge__(self, value: object) -> ResultComparison:
        """."""
        return self._create_operation(">=", value)

    def is_(self, value: object) -> ResultComparison:
        """."""
        return self._create_operation("IS", value)

    def is_not(self, value: object) -> ResultComparison:
        """."""
        return self._create_operation("IS NOT", value)

    def in_(self, value: Sequence | Query) -> ResultComparison:
        """Realization in operation."""
        if not value:
            msg = "Value could not be empty"
            raise ValueError(msg)
        if isinstance(value, Sequence):
            if isinstance(value[0], Number):
                value = "(" + ", ".join(str(i) for i in value) + ")"
            elif isinstance(value[0], str):
                value = "(" + ", ".join(f"'{i}'"for i in value) + ")"
        else:
            # if Query
            value = "({})".format(str(value))
        return self._create_operation("IN", value, in_operation=True)

    def not_in(self, value: Sequence | Query) -> str:
        """Realization in operation."""
        query = self.in_(value)
        return query.replace("IN", "NOT IN")

    def __getattr__(self, name: TFunc | str) -> Callable[[], SQLBase]:
        """Realize sql funcs."""
        FUNC_NAMES = {
            "avg",
            "sum",
            "min",
            "max",
            "count",
            "lower",
            "upper",
            "length",
        }
        if name not in FUNC_NAMES:
            msg = "{} {} {}".format(
                self.__class__.__name__,
                "object has no attribute",
                name,
            )
            raise AttributeError(msg)
        func_cls = getattr(func, name)
        return lambda: func_cls(self)

    def __str__(self) -> str:
        """Return column full name."""
        return self._full_column_name

    @property
    def row_name(self) -> str:
        """Row name."""
        return self._column_name

    def __and__(self, value: Any) -> None:
        """Stub."""
        msg = (
            "Use parentheses. Like "
            "(table.c.name = '1') & (table.c.salary.is_(None))"
        )
        raise ValueError(msg)

    __or__ = __and__

    __rand__ = __and__

    __ror__ = __and__


class Columns:

    __slots__ = ("_table",)

    def __init__(self, table: Table) -> None:
        """Initialize."""
        self._table = table

    def __getattr__(self, name: str) -> Column:
        """Return column object."""
        if name not in self._table.column_names:
            msg = "{} {} {}".format(
                self.__class__.__name__,
                "object has no attribute",
                name,
            )
            raise AttributeError(msg)
        return Column(self._table.name, name)



"""Module contains column types for create table."""
from __future__ import annotations

import json
from abc import abstractmethod
from collections.abc import Mapping
from dataclasses import asdict
from dataclasses import fields
from dataclasses import is_dataclass
from datetime import date
from datetime import datetime
from datetime import time
from decimal import Decimal
from enum import Enum
from fractions import Fraction
from functools import singledispatchmethod
from numbers import Number
from typing import Any
from typing import Literal
from typing import Protocol
from typing import TypedDict
from typing import TypeVar
from typing import Union

from typing_extensions import NotRequired
from typing_extensions import TypeAlias
from typing_extensions import Unpack

from .enumcls import DeleteAction
from .enumcls import UpdateAction


__all__ = (
    "BaseType",
    "Integer",
    "DataClassJson",
    "Real",
    "Text",
    "Blob",
    "ForeignKey",
    "Json",
    "DateTime",
    "Date",
    "Time",
    "EnumName",
    "EnumValue",
    "EnumHide",
)


TNumberType = TypeVar("TNumberType", bound="Number")
TDateTimeDBFormat: TypeAlias = Literal["timestamp", "ISO"]
TDateDBFormat: TypeAlias = Literal["timestamp", "ISO", "ordinal"]


class TColumnType(Protocol):
    """Column type."""

    __slots__ = ()

    def __str__(self) -> str:
        """String view."""
        ...

    def to_db(self, value: Any) -> Any:
        """Serialize data to db type."""
        ...

    def to_python(self, value: Any) -> Any:
        """Serialize data to python type."""


class TBaseTypeKwargs(TypedDict):
    """Base type kwargs."""

    primary_key: NotRequired[bool]
    unique: NotRequired[bool]
    nullable: NotRequired[bool]


class ORMColumnProtocol(Protocol):
    """ORM column serializers."""

    __slots__ = ()

    def to_db(self, value: Any) -> Any:
        ...

    def to_python(self, value: Any) -> Any:
        ...


class ColumnString:
    """Columns string."""

    __slots__ = ("data",)

    def __init__(self, seq: str) -> None:
        self.data = seq

    def __repr__(self) -> str: return repr(self.data)
    def __int__(self) -> int: return int(self.data)
    def __float__(self) -> float: return float(self.data)
    def __complex__(self) -> complex: return complex(self.data)
    def __hash__(self) -> int: return hash(self.data)
    def __len__(self) -> int: return len(self.data)

    @abstractmethod
    def __str__(self) -> str:
        """String view."""
        ...


class BaseType(ColumnString):
    """Base column type."""

    __slots__ = (
        "default",
        "primary_key",
        "unique",
        "nullable",
    )

    def __init__(
        self,
        str_type: str,
        default: Any | None = None,
        *,
        primary_key: bool = False,
        unique: bool = False,
        nullable: bool = True,
    ) -> None:
        """Initialize base params for abstract type.

        Args:
            default (int | str | None): default value. Defaults to None.
            primary_key (bool): is that primary key. Defaults to False.
            unique (bool): is that unique. Defaults to False.
            nullable (bool): maybe null. Defaults to False.

        """
        super().__init__(str_type)
        self.default = default
        self.primary_key = primary_key
        self.unique = unique
        self.nullable = nullable

    def __str__(self) -> str:
        """Create string column type."""
        sql_column_type = self.data
        if self.default:
            if isinstance(self.default, str):
                sql_column_type += f" DEFAULT '{self.default}' "
            else:
                sql_column_type += f" DEFAULT {self.default} "
        if self.primary_key:
            sql_column_type += " PRIMARY KEY "
            self.nullable = False
        if not self.nullable:
            sql_column_type += " NOT NULL "
        if self.unique:
            sql_column_type += " UNIQUE "
        return sql_column_type


class Integer(BaseType):
    """Integer type."""

    __slots__ = ("autoincrement",)

    def __init__(
        self,
        default: int | None = None,
        *,
        autoincrement: bool = False,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        """Initialize base params for INTEGER.

        Args:
            default (int | str | None): default value. Defaults to None.
            primary_key (bool): is that primary key. Defaults to False.
            unique (bool): is that unique. Defaults to False.
            nullable (bool): maybe null. Defaults to False.
            autoincrement (bool): on autoincrement. Defaults to False.

        """
        if default is not None and not isinstance(default, int):
            msg = "Incorrect type for default value."
            raise TypeError(msg)

        self.autoincrement = autoincrement
        super().__init__(
            "INTEGER",
            default=default,
            **kwargs,
        )

    def __str__(self) -> str:
        """Create string column type."""
        sql_column_type = super().__str__()
        if self.autoincrement:
            sql_column_type = sql_column_type.replace(
                "PRIMARY KEY",
                "PRIMARY KEY AUTOINCREMENT",
            )
        return sql_column_type

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type value can be only number"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> int:
        """Serialize python obj to db data."""
        return int(value)

    @to_db.register
    def _(self, value: int) -> int:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: float) -> int:
        """Serialize python obj to db data."""
        return int(value)

    @to_db.register
    def _(self, value: Decimal) -> int:
        """Serialize python obj to db data."""
        return int(value)

    @to_db.register
    def _(self, value: Fraction) -> int:
        """Serialize python obj to db data."""
        return int(value)

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: int) -> int:
        """Serialize python obj to db data."""
        return value

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value


class Real(BaseType):
    """REAL column type."""

    __slots__ = ()

    def __init__(
        self,
        default: float | int | None = None,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        """Initialize base params for REAL.

        Args:
            default (int | str | None): default value. Defaults to None.
            primary_key (bool): is that primary key. Defaults to False.
            unique (bool): is that unique. Defaults to False.
            nullable (bool): maybe null. Defaults to False.
            autoincrement (bool): on autoincrement. Defaults to False.

        """
        if default is not None and not isinstance(default, Number):
            msg = "Incorrect type for default value."
            raise TypeError(msg)

        super().__init__(
            "REAL",
            default=default,
            **kwargs,
        )

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type, value can be only number"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: float) -> float:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: int) -> int:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: Decimal) -> Decimal:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: Fraction) -> Fraction:
        """Serialize python obj to db data."""
        return value

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type, value can be only number"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: float) -> float:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: int) -> int:
        """Serialize db data to python obj."""
        return value


class Text(BaseType):
    """TEXT column type."""

    __slots__ = ()

    def __init__(
        self,
        default: str | None = None,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        """Initialize base params for TEXT.

        Args:
            default (int | str | None): default value. Defaults to None.
            primary_key (bool): is that primary key. Defaults to False.
            unique (bool): is that unique. Defaults to False.
            nullable (bool): maybe null. Defaults to False.
            autoincrement (bool): on autoincrement. Defaults to False.

        """
        if default is not None and not isinstance(default, str):
            msg = "Incorrect type for default value."
            raise TypeError(msg)

        super().__init__(
            "TEXT",
            default=default,
            **kwargs,
        )

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type, value can be only str"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> str:
        """Serialize python obj to db data."""
        return value

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type, can be only str"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: str) -> str:
        """Serialize db data to python obj."""
        return value


class Blob(BaseType):
    """BLOB column type."""

    __slots__ = ()

    def __init__(
        self,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        """Initialize base params for BLOB.

        Args:
            default (int | str | None): default value. Defaults to None.
            primary_key (bool): is that primary key. Defaults to False.
            unique (bool): is that unique. Defaults to False.
            nullable (bool): maybe null. Defaults to False.
            autoincrement (bool): on autoincrement. Defaults to False.

        """
        super().__init__(
            "BLOB",
            **kwargs,
        )

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type, value can be only bytes"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: bytes) -> bytes:
        """Serialize python obj to db data."""
        return value

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type, can be only str"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: bytes) -> bytes:
        """Serialize db data to python obj."""
        return value


class ForeignKey(BaseType):
    """Foreign key constraint."""

    __slots__ = (
        "column",
        "second_table",
        "reference_column",
        "on_delete",
        "on_update",
    )

    def __init__(
        self,
        column: str,
        second_table: str,
        reference_column: str,
        on_delete: DeleteAction | None = None,
        on_update: UpdateAction | None = None,
    ) -> None:
        """Initialize."""
        self.column = column
        self.second_table = second_table.lower()
        self.reference_column = reference_column
        self.on_delete = DeleteAction(on_delete) if on_delete else on_delete
        self.on_update = UpdateAction(on_update) if on_update else on_update
        super().__init__("FOREIGN KEY(`{}`) REFERENCES `{}`(`{}`)")

    def __call__(self) -> str:
        """Create command."""
        stmt = self.data.format(
            self.column,
            self.second_table,
            self.reference_column,
        )
        if self.on_delete:
            stmt += f" ON DELETE {self.on_delete.value}"
        if self.on_update:
            stmt += f" ON UPDATE {self.on_update.value}"
        return stmt

    def __str__(self) -> str:
        return self()


class Json(BaseType):
    """TEXT column type with dict conversion."""

    __slots__ = (
        "_json_module",
    )

    def __init__(
        self,
        default: str | None = None,
        *,
        json_module: Any = json,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        super().__init__(
            "TEXT",
            default,
            **kwargs,
        )
        self._json_module = json_module

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type, value can be only str or dict"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> str:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: Mapping) -> str:
        """Serialize python obj to db data."""
        return self._json_module.dumps(value)

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type, can be only str"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: str) -> dict:
        """Serialize db data to python obj."""
        return self._json_module.loads(value)


class DataClassJson(BaseType):
    """Text column."""

    __slots__ = (
        "_json_module",
        "_data_class",
    )

    def __init__(
        self,
        default: str | None = None,
        *,
        data_class: Any,
        json_module: Any = json,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        super().__init__(
            "TEXT",
            default,
            **kwargs,
        )
        if is_dataclass(data_class) is False:
            msg = "data_class is not a data class"
            raise TypeError(msg)
        self._data_class = data_class
        self._json_module = json_module

    @singledispatchmethod
    def to_db(self, value: Any) -> str:
        """Serialize python obj to db data."""
        if is_dataclass(value):
            return self._json_module.dumps(
                asdict(value),  # type: ignore[arg-type]
            )
        msg = "Unknown type"
        raise TypeError(msg)

    @to_db.register  # type: ignore[arg-type]
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> str:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: Mapping) -> str:
        """Serialize python obj to db data."""
        return self._json_module.dumps(value)

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: str) -> Any:
        """Serialize db data to python obj."""
        data = self._json_module.loads(value)
        if not data:
            data = {
                field.name: None
                for field in fields(self._data_class)
            }
        return self._data_class(**data)


class DateTime(BaseType):
    """TEXT or REAL column type with dict conversion."""

    __slots__ = ("_datetime_db_format",)

    def __init__(
        self,
        default: str | None = None,
        *,
        datetime_db_format: TDateTimeDBFormat = "ISO",
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        db_type = "TEXT"

        if datetime_db_format not in {"ISO", "timestamp"}:
            msg = "Unknown format for datetime_db_format"
            raise ValueError(msg)

        self._datetime_db_format = datetime_db_format
        if datetime_db_format == "timestamp":
            db_type = "REAL"

        super().__init__(
            db_type,
            default,
            **kwargs,
        )

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> str:
        """Serialize python obj to db data."""
        if self._datetime_db_format != "ISO":
            msg = "Unknown type"
            raise TypeError(msg)
        return value

    @to_db.register
    def _(self, value: float) -> float:
        """Serialize python obj to db data."""
        if self._datetime_db_format != "timestamp":
            msg = "Unknown type"
            raise TypeError(msg)
        return value

    @to_db.register
    def _(self, value: datetime) -> Union[float, str]:
        """Serialize python obj to db data."""
        if self._datetime_db_format == "timestamp":
            return value.timestamp()
        return value.isoformat()

    @to_db.register
    def _(self, value: date) -> Union[float, str]:
        """Serialize python obj to db data."""
        datetime_value = datetime.combine(value, datetime.min.time())
        if self._datetime_db_format == "timestamp":
            return datetime_value.timestamp()
        return datetime_value.isoformat()

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type, can be only str"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: str) -> datetime:
        """Serialize db data to python obj."""
        if self._datetime_db_format != "ISO":
            msg = "Unknown type"
            raise TypeError(msg)

        return datetime.fromisoformat(value)

    @to_python.register
    def _(self, value: float) -> datetime:
        """Serialize db data to python obj."""
        if self._datetime_db_format != "timestamp":
            msg = "Unknown type"
            raise TypeError(msg)

        return datetime.fromtimestamp(value)


class Date(BaseType):
    """TEXT or REAL or INTEGER column type with dict conversion."""

    __slots__ = ("_date_db_format",)

    def __init__(
        self,
        default: str | None = None,
        *,
        date_db_format: TDateDBFormat = "ISO",
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        db_type = "TEXT"
        self._date_db_format = date_db_format

        if date_db_format not in {"timestamp", "ISO", "ordinal"}:
            msg = "Unknown name for date_db_format"
            raise ValueError(msg)

        if date_db_format == "timestamp":
            db_type = "REAL"

        if date_db_format == "ordinal":
            db_type = "INTEGER"

        super().__init__(
            db_type,
            default,
            **kwargs,
        )

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> str:
        """Serialize python obj to db data."""
        if self._date_db_format != "ISO":
            msg = "Unknown type"
            raise TypeError(msg)
        return value

    @to_db.register
    def _(self, value: float) -> float:
        """Serialize python obj to db data."""
        if self._date_db_format != "timestamp":
            msg = "Unknown type"
            raise TypeError(msg)
        return value

    @to_db.register
    def _(self, value: datetime) -> Union[float, str, int]:
        """Serialize python obj to db data."""
        if self._date_db_format == "timestamp":
            return value.timestamp()
        if self._date_db_format == "ordinal":
            return value.toordinal()
        return value.date().isoformat()

    @to_db.register
    def _(self, value: date) -> Union[float, str, int]:
        """Serialize python obj to db data."""
        if self._date_db_format == "timestamp":
            datetime_value = datetime.combine(value, datetime.min.time())
            return datetime_value.timestamp()
        if self._date_db_format == "ordinal":
            return value.toordinal()
        return value.isoformat()

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: float) -> date:
        """Serialize db data to python obj."""
        if self._date_db_format != "timestamp":
            msg = "Unknown type"
            raise TypeError(msg)
        return datetime.fromtimestamp(value).date()

    @to_python.register
    def _(self, value: str) -> date:
        """Serialize db data to python obj."""
        if self._date_db_format != "ISO":
            msg = "Unknown type"
            raise TypeError(msg)

        return datetime.fromisoformat(value).date()

    @to_python.register
    def _(self, value: int) -> date:
        """Serialize db data to python obj."""
        if self._date_db_format != "ordinal":
            msg = "Unknown type"
            raise TypeError(msg)

        return datetime.fromordinal(value).date()


class Time(BaseType):
    """TEXT column type with dict conversion."""

    __slots__ = ()

    def __init__(
        self,
        default: str | None = None,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        super().__init__(
            "TEXT",
            default,
            **kwargs,
        )

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type, value can be only str or dict"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> str:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: time) -> str:
        """Serialize python obj to db data."""
        return value.isoformat()

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type, can be only str"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: str) -> time:
        """Serialize db data to python obj."""
        return time.fromisoformat(value)


class EnumName(BaseType):
    """"TEXT type that stores the name from an Enum object."""

    __slots__ = (
        "_enum_cls",
        "_enum_names",
    )

    def __init__(
        self,
        enum_cls: type[Enum],
        *,
        default: str | None = None,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        super().__init__(
            "TEXT",
            default,
            **kwargs,
        )
        self._enum_cls = enum_cls
        self._enum_names = {item.name for item in enum_cls}

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type, value can be only str or dict"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> str:
        """Serialize python obj to db data."""
        if value in self._enum_names:
            return value
        msg = f"Enum do not have a '{value}' name"
        raise ValueError(msg)

    @to_db.register
    def _(self, value: Enum) -> str:
        """Serialize python obj to db data."""
        return value.name

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type, can be only str"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: str) -> Any:
        """Serialize db data to python obj."""
        return getattr(self._enum_cls, value)


class EnumValue(BaseType):
    """"TEXT type that stores the value from an Enum object."""

    __slots__ = (
        "_enum_cls",
        "_enum_values",
    )

    def __init__(
        self,
        enum_cls: type[Enum],
        *,
        default: Any = None,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        super().__init__(
            "TEXT",
            default,
            **kwargs,
        )
        self._enum_cls = enum_cls
        self._enum_values = {item.value for item in enum_cls}

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        if value in self._enum_values:
            return value
        msg = f"Enum do not have a '{value}' value"
        raise ValueError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: Enum) -> Any:
        """Serialize python obj to db data."""
        return value.value

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        return self._enum_cls(value)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value


class EnumHide(BaseType):
    """"
    TEXT type that stores the name from an Enum object.
    And return only value without enum obj.
    """

    __slots__ = (
        "_enum_cls",
        "_enum_names",
    )

    def __init__(
        self,
        enum_cls: type[Enum],
        *,
        default: str | None = None,
        **kwargs: Unpack[TBaseTypeKwargs],
    ) -> None:
        super().__init__(
            "TEXT",
            default,
            **kwargs,
        )
        self._enum_cls = enum_cls
        self._enum_names = {item.name for item in enum_cls}

    @singledispatchmethod
    def to_db(self, value: Any) -> Any:
        """Serialize python obj to db data."""
        msg = "Unknown type"
        raise TypeError(msg)

    @to_db.register
    def _(self, value: None) -> None:
        """Serialize python obj to db data."""
        return value

    @to_db.register
    def _(self, value: str) -> str:
        """Serialize python obj to db data."""
        if value in self._enum_names:
            return value
        msg = f"Enum do not have a '{value}' name"
        raise ValueError(msg)

    @to_db.register
    def _(self, value: Enum) -> str:
        """Serialize python obj to db data."""
        return value.name

    @singledispatchmethod
    def to_python(self, value: Any) -> Any:
        """Serialize db data to python obj."""
        msg = "Unknown type"
        raise TypeError(msg)

    @to_python.register
    def _(self, value: None) -> None:
        """Serialize db data to python obj."""
        return value

    @to_python.register
    def _(self, value: str) -> Any:
        """Serialize db data to python obj."""
        return getattr(self._enum_cls, value).value



"""Module contain components for work with db table."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import Iterator
from typing import Literal
from typing import Type
from typing import TypeVar
from typing import cast

from ..enumcls import ResultFetch
from ..operations import Delete
from ..operations import Insert
from ..operations import Query
from ..operations import Select
from ..operations import Update
from ..orm.utils import contain_relation_objects
from ..orm.utils import process_add_relation_objects
from ..orm.utils import refresh_old_obj_by_new
from ..rows import RowDict
from ..rows import make_row_data_cls
from .column import Columns


if TYPE_CHECKING:
    import sqlite3

    from ..db import DB
    from ..orm.typings import TModel


T = TypeVar("T")


__all__ = (
    "Table",
)


class Table(Generic[T]):
    """Component for work with table."""

    __slots__ = (
        "_name",
        "select",
        "insert",
        "delete",
        "update",
        "use_datacls",
        "c",
        "columns",
        "db",
        "column_names",
        "primary_keys",
        "row_cls",
        "_query_obj",
        "execute",
    )

    def __init__(
        self,
        name: str | None = None,
        *,
        use_datacls: bool = False,
        row_cls: type[T] | None = None,
        query_cls: Type[Query] | None = None,
        insert_cls: Type[Insert[T]] | None = None,
        delete_cls: Type[Delete] | None = None,
        update_cls: Type[Update] | None = None,
    ) -> None:
        """Initialize."""
        if name is None:
            msg = "Table name do not be None."
            raise ValueError(msg)

        self._name = name
        self.primary_keys: tuple[str, ...] = ()
        self.use_datacls = use_datacls
        self.row_cls = cast(T, row_cls or RowDict)

        # Operations
        self._query_obj = query_cls if query_cls else Query
        self.select: Select[T] = Select(self)
        self.insert = insert_cls(self) if insert_cls else Insert[T](self)
        self.delete = delete_cls(self) if delete_cls else Delete(self)
        self.update = update_cls(self) if update_cls else Update(self)

        # Sugar
        # self.add = self.insert

        self.c = Columns(self)
        self.columns = self.c

    @property
    def query(self) -> Query[T]:
        """Return new query object for this table."""
        return self._query_obj(self)

    @property
    def name(self) -> str:
        """Return table name."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Return table name."""
        self._name = value

    @property
    def cursor(self) -> sqlite3.Cursor:
        """Shortcut for cursor."""
        return self.db.connect.cursor()

    # @property
    # def execute(
    #     self,
    # ) -> Callable[..., list[tuple] | None]:
    #     """Shortcut for execute.

    #     Args:
    #         query (str): sql query
    #         parameters (MutableMapping | Sequence): data for executing.
    #         Defaults to ().
    #         many (bool): flag for executemany operation. Defaults to False.
    #         size (int | None): size for fetchmany operation. Defaults to None.
    #         result (ResultFetch | None): enum for fetch func. Defaults to None.

    #     Returns:
    #         list[Any] or None

    #     """
    #     return self.db.execute

    def _get_column_names(self) -> tuple[str, ...]:
        """Fetch table column name."""
        stmt = "SELECT name, pk FROM PRAGMA_TABLE_INFO('{}');".format(
            self.name,
        )
        result = self.db.execute(stmt, result=ResultFetch.fetchall)
        names = []
        primary_keys: list[str] = []

        for row in result:
            names.append(row[0])
            if row[1] == 1:
                primary_keys.append(row[0])

        self.primary_keys = tuple(primary_keys)
        return tuple(names)

    # @cached_property
    # def id_exist(self) -> bool:
    #     """Check exist id column."""
    #     return "id" in self.column_names

    def all(self) -> list[T]:
        """Get all rows from table."""
        return self.select()

    def __iter__(self) -> Iterator[T]:
        """Iterate through the row list."""
        return self.select().__iter__()

    def __getitem__(self, index: int) -> T:
        """Get row item by index in list."""
        return self.query().all()[index]

    def get(self, **filter_by: Any) -> T | None:
        """Get one row by filter."""
        return self.query().where(**filter_by).limit(1).first()

    def drop(self, *, init_tables: bool = True) -> None:
        """Drop this table."""
        self.db.execute(f"DROP TABLE IF EXISTS {self.name}")
        # TODO(stickking): What ?!?! replace in db.execute
        # 0000
        if init_tables:
            self.db.initialize_tables()

    def add(
        self,
        *objects: dict | Any,
        returning: bool = False,
    ) -> None | T:
        """Add objects in table."""
        data = []
        objects_with_relation: list[TModel] = []
        insert_obj = None

        for obj in objects:
            if isinstance(obj, dict):
                data.append(obj)
                continue

            obj.table = self
            process_add_relation_objects(obj, ref_type="RelationForeignKey")

            if contain_relation_objects(obj):
                objects_with_relation.append(obj)
                continue

            data.append(obj.get_row_data_as_dict())

        if data:
            insert_obj = self.insert(
                data,
                returning=cast(Literal[True, False], returning),
            )

        new_obj = None
        # too slow
        for obj in objects_with_relation:
            new_obj = self.insert(
                obj.get_row_data_as_dict(),
                returning=True,
            )
            refresh_old_obj_by_new(self, obj, new_obj)
            process_add_relation_objects(obj)

        return insert_obj or new_obj

    def __repr__(self) -> str:
        """Repr view."""
        return f"<{self.__class__.__name__}: {self.name.title()}>"

    def __call__(self, db: DB) -> None:
        """Prepare table obj."""
        self.db = db
        self.execute = db.execute

        self.column_names = self._get_column_names()
        if self.use_datacls and self.row_cls == RowDict:
            self.row_cls = cast(
                T,
                make_row_data_cls(
                    self.name,
                    self.column_names,
                ),
            )



"""Module contains row classes."""
from __future__ import annotations

import sys
from abc import ABC
from abc import abstractmethod
from dataclasses import _process_class  # type: ignore
from dataclasses import field
from dataclasses import make_dataclass
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Sequence
from typing import TypeVar


if TYPE_CHECKING:
    from .table.table import Table


TRow = TypeVar("TRow", bound="ABCRow")


__all__ = (
    "ABCRow",
    "RowDict",
    "_RowDataClsMixin",
    "make_row_data_cls",
)


class ABCRow(ABC):
    """Abstract row interface."""

    table: Table
    changed_columns: set

    @property
    @abstractmethod
    def not_changed_column_values(self) -> dict[str, Any]:
        """Fetch not changed column name with value like dict."""
        ...

    @property
    @abstractmethod
    def changed_column_values(self) -> dict[str, Any]:
        """Fetch changed column name with value like dict."""
        ...

    def delete(self) -> None:
        """Delete this row from db."""
        self.table.delete(**self.not_changed_column_values)

    def change(self) -> None:
        """Update this row."""
        if not self.changed_columns:
            return
        self.table.update(
            self.changed_column_values,
            **self.not_changed_column_values,
        )
        self.changed_columns = set()


class _BaseRowDataClsMixin(ABCRow):
    """Mixin for realize change control in row."""

    def __repr__(self: Any) -> str:
        """View string by obj."""
        columns = ", ".join(
            f"{atr_name}={getattr(self, atr_name)}"
            for atr_name in self.table.column_names
        )
        return f"{self.__class__.__name__}({columns})"


class _RowDataClsMixin(_BaseRowDataClsMixin):
    """Row data cls mixin."""

    def __setattr__(self, name: str, value: Any) -> None:
        """Check changed attribute for updating and deleting row."""
        if (
            hasattr(self, "changed_columns") and
            self.changed_columns is not None and
            self.table is not None and
            name.startswith("_") is False and
            name in self.table.column_names
        ):
            old_value = getattr(self, name)
            super().__setattr__(name, value)
            if value != old_value:
                self.changed_columns.add(name)
            return
        super().__setattr__(name, value)

    @property
    def not_changed_column_values(self) -> dict[str, Any]:
        """Fetch not changed column name with value like dict."""
        not_change_column = set(self.table.column_names) - self.changed_columns
        return {
            name: getattr(self, name)
            for name in self.table.column_names
            if name in not_change_column
        }

    @property
    def changed_column_values(self) -> dict[str, Any]:
        """Fetch changed column name with value like dict."""
        return {
            name: getattr(self, name)
            for name in self.table.column_names
            if name in self.changed_columns
        }


class RowDict(ABCRow, dict):
    """DB row like a dict."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize."""
        self.table = kwargs.pop("table")
        self.changed_columns = set()
        if self.table is None:
            msg = "missing 1 required named argument: 'table'"
            raise TypeError(msg)
        super().__init__(*args, **kwargs)

    @property
    def not_changed_column_values(self) -> dict[str, Any]:
        """Fetch not changed column name with value like dict."""
        not_change_column = set(self.table.column_names) - self.changed_columns
        return {
            key: value
            for key, value in self.items()
            if key in not_change_column
        }

    @property
    def changed_column_values(self) -> dict[str, Any]:
        """Fetch changed column name with value like dict."""
        return {
            key: value
            for key, value in self.items()
            if key in self.changed_columns
        }

    def __setitem__(self, key: str, value: int | str | bool) -> None:
        """Check changes columns."""
        if self[key] != value and key in self.table.column_names:
            self.changed_columns.add(key)
        super().__setitem__(key, value)


def make_row_data_cls(
    table_name: str,
    column_names: Sequence[str],
    bases: Sequence[type[Any]] | None = None,
    *,
    default_none: bool = True,
    create_orm_model: bool = False,
) -> type[Any]:
    """Create data cls row for the transmitted table."""
    attributes: list[tuple[str, Any, Any]] = []
    init = True

    if bases is None:
        bases = [_RowDataClsMixin]
    else:
        bases = [_RowDataClsMixin, *bases]

    attributes = [
        (atr, Any, field(default=None))
        for atr in column_names
    ]
    attributes.extend([
        ("changed_columns", set, field(default_factory=lambda: set())),
        ("table", Any, field(default=None))
    ])

    return make_dataclass(
        f"Row{table_name}DataClass",
        attributes,
        repr=False,
        bases=(*bases,),
        init=init,
    )

    # data_cls.__repr__ = repr

    # return type(
    #     f"Row{table.name.title()}DataClass",
    #     (data_cls, _RowDataClsMixin),
    #     {},
    #     # {"__slots__": data_cls.__slots__},
    # )


def create_result_row(columns_name: Iterable[str]) -> type[Any]:
    """Create result row cls."""
    columns = list(columns_name)
    columns.append("table")
    return make_dataclass(
        "ResultRow",
        columns,
        frozen=True,
    )


def dataclass_row(  # noqa: PLR0913
    cls: None | type = None,
    /,
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
    **kwargs: Any,
) -> type | Callable[[type], type]:
    """Make custom row dataclass with mixin, repr
    and arguments: 'table', 'changed_columns'.
    """
    python_version = sys.version_info

    if python_version.major == 3 and python_version.minor >= 10:
        kwargs["match_args"] = kwargs.get("match_args", True)
        kwargs["kw_only"] = kwargs.get("kw_only", False)
        kwargs["slots"] = kwargs.get("slots", False)

    if python_version.major == 3 and python_version.minor >= 11:
        kwargs["weakref_slot"] = kwargs.get("weakref_slot", False)

    def wrap(cls: type) -> type:
        cls.__annotations__["table"] = Any
        cls.__annotations__["changed_columns"] = set
        cls.table = field(default=None)  # type: ignore
        cls.changed_columns = field(  # type: ignore
            default_factory=lambda: set(),
        )

        # TODO (stickking): Remove _process_class and take wrap
        # 0000
        new_cls = _process_class(
            cls,
            init,
            repr,
            eq,
            order,
            unsafe_hash,
            frozen,
            **kwargs,
        )
        # TODO (stickking): Why dict if slots?!?!
        # 0000
        return type(
            new_cls.__name__,
            (cls, _RowDataClsMixin),
            {} if hasattr(cls, "__dict__")
            else {"__slots__": new_cls.__slots__},
        )

    # See if we're being called as @dataclass or @dataclass().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @dataclass without parens.
    cls = wrap(cls)

    if repr is False:
        cls.__repr__ = cls.__str__  # type: ignore

    return cls



"""Module contain orm typings."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from typing import Protocol


if TYPE_CHECKING:
    from collections import defaultdict

    from ..table import Table


class TModel(Protocol):
    """Model protocol."""

    __table_name__: ClassVar[str]
    __relation_fields__: ClassVar[tuple[str, ...]]
    __column_fields__: ClassVar[tuple[str, ...]]

    table: Table
    changed_columns: set[str]
    _relation_events: defaultdict[str, list]
    not_changed_column_values: dict[str, Any]
    changed_column_values: dict[str, Any]

    def get_row_data_as_dict(self) -> dict:
        """Return row data like dict."""
        ...

    def change(self) -> None:
        """Update this row."""
        ...



"""Module contains functions that generate an ORM model of the database."""
from __future__ import annotations

import datetime
from collections import defaultdict
from typing import Any
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Type
from typing import TypedDict
from typing import Union

from typing_extensions import NotRequired
from typing_extensions import TypeAlias
from typing_extensions import get_args
from typing_extensions import get_origin
from typing_extensions import get_type_hints
from typing_extensions import get_annotations
from typing_extensions import Format

from ..column_types import BaseType
from ..column_types import Blob
from ..column_types import Date
from ..column_types import DateTime
from ..column_types import ForeignKey
from ..column_types import Integer
from ..column_types import Real
from ..column_types import Text
from ..column_types import Time
from ..orm import MColumn
from ..orm import Relation
from ..orm import TColumn
from ..rows import _BaseRowDataClsMixin
from .utils import process_add_relation_objects


TForeignKeys: TypeAlias = List[ForeignKey]
TTableColumns: TypeAlias = Dict[str, BaseType]
TModelClass: TypeAlias = Type[object]


class TableData(TypedDict):
    """Data for creating the table."""

    table_name: str
    columns: TTableColumns
    foreign_keys: NotRequired[TForeignKeys]


TYPE_MAPPER = {
    # Optional types
    Union[int, None]: Integer,
    Union[str, None]: Text,
    Union[bytes, None]: Blob,
    Union[float, None]: Real,
    Union[datetime.datetime, None]: DateTime,
    Union[datetime.date, None]: Date,
    Union[datetime.time, None]: Time,
    # Not null types
    int: lambda: Integer(nullable=False),
    str: lambda: Text(nullable=False),
    bytes: lambda: Blob(nullable=False),
    float: lambda: Real(nullable=False),
    datetime.datetime: lambda: DateTime(nullable=False),
    datetime.date: lambda: Date(nullable=False),
    datetime.time: lambda: Time(nullable=False),
}


class _RowORMModelMixin(_BaseRowDataClsMixin):
    """ORM row mixin."""

    __table_name__: ClassVar[str]
    __relation_fields__: ClassVar[tuple[str, ...]]
    __column_fields__: ClassVar[tuple[str, ...]]

    def __init__(self, **kwargs: Any) -> None:
        """Initialize orm object."""
        super().__setattr__("_is_init", True)

        self.table = kwargs.pop("table", None)
        self.changed_columns = set()
        self._relation_events: defaultdict[str, list] = (
            defaultdict(list)
        )

        column_fields = set(self.__column_fields__)
        column_fields.update(self.__relation_fields__)

        for key, value in kwargs.items():
            if key not in column_fields:
                continue
            setattr(self, key, value)
            column_fields.remove(key)

        for name in column_fields:
            object.__setattr__(self, f"_column_data_{name}_", None)

        self.orm_obj = True
        self._is_init = False

    def get_row_data_as_dict(self) -> dict:
        """Return row data like dict."""
        return {
            name: getattr(self, f"_column_data_{name}_")
            for name in self.table.column_names
            if getattr(self, f"_column_data_{name}_") is not None
        }

    def __setattr__(self, name: str, value: Any) -> None:
        """Check changed attribute for updating and deleting row."""
        if self._is_init:
            super().__setattr__(name, value)
            return

        if (
            hasattr(self, "changed_columns") and
            self.changed_columns is not None and
            self.table is not None and
            name.startswith("_") is False and
            name in self.table.column_names
        ):
            old_value = getattr(self, name)
            super().__setattr__(name, value)
            if value != old_value:
                self.changed_columns.add(name)
            return
        super().__setattr__(name, value)

    @property
    def not_changed_column_values(self) -> dict[str, Any]:
        """Fetch not changed column name with value like dict."""
        not_change_column = set(self.table.column_names) - self.changed_columns
        return {
            name: getattr(self, f"_column_data_{name}_")
            for name in self.table.column_names
            if name in not_change_column
        }

    @property
    def changed_column_values(self) -> dict[str, Any]:
        """Fetch changed column name with value like dict."""
        return {
            name: getattr(self, f"_column_data_{name}_")
            for name in self.table.column_names
            if name in self.changed_columns
        }

    def change(self) -> None:
        """Update this row."""
        process_add_relation_objects(self)
        super().change()

    def delete(self) -> None:
        """Delete this row from db."""
        self.table.delete(**self.not_changed_column_values)


def _get_table_columns(
    model_cls: TModelClass,
    table_columns: TTableColumns,
    foreign_keys: TForeignKeys,
    relation_names: list[str],
) -> None:
    """Get table columns from cls."""
    try:
        annotations: dict[str, TColumn] = get_type_hints(
            model_cls,
            format=Format.STRING,
        )
    except TypeError:
        # Python 3.12
        annotations = get_annotations(model_cls)
        for type_ in annotations.values():
            if isinstance(type_, str) and "TColumn" in type_:
                msg = "Remove 'from __future__ import annotations'"
                raise ImportError(msg)

    parent_cls, *_ = model_cls.__bases__

    if parent_cls is not object:
        for parent_cls in model_cls.__bases__:
            _get_table_columns(
                parent_cls,
                table_columns,
                foreign_keys,
                relation_names,
            )

    for key, value in model_cls.__dict__.items():

        if isinstance(value, Relation):
            relation_names.append(key)

        if isinstance(value, ForeignKey):
            foreign_keys.append(value)
            relation_names.append(key)
            continue

        if (
            key.startswith("__") or
            isinstance(value, MColumn) is False or
            key in table_columns  # exists in child cls
        ):
            continue

        table_columns[key] = value.column_type

    for key, value in annotations.items():

        # exists in cls attr
        if key in table_columns:
            continue

        real_cls = get_origin(value)
        if real_cls is not TColumn:
            continue

        python_type, = get_args(value)

        if python_type not in TYPE_MAPPER:
            continue

        # get db type
        db_type_cls = TYPE_MAPPER.get(python_type)

        if db_type_cls is None:
            msg = "Unknown type"
            raise TypeError(msg)

        db_type = db_type_cls()  # type: ignore[operator]

        # set cls column
        setattr(model_cls, key, MColumn(db_type))
        model_cls.__dict__[key].__set_name__(
            model_cls,
            key,
        )

        table_columns[key] = db_type


def create_table_and_data_cls_row(
    model_cls: TModelClass,
) -> tuple[TableData, type[Any]]:
    """Create table by model class and create row class."""

    table_columns: TTableColumns = {}
    foreign_keys: TForeignKeys = []
    relation_names: list[str] = []
    _get_table_columns(model_cls, table_columns, foreign_keys, relation_names)

    table_name = model_cls.__name__.lower()

    # if hasattr(model_cls, "__table_name__"):
    #     table_name = model_cls.__table_name__

    table_data = TableData(
        table_name=table_name,
        columns=table_columns,
    )

    if foreign_keys:
        table_data["foreign_keys"] = foreign_keys

    # row_cls = make_row_data_cls(
    #     table_name,
    #     list(table_columns.keys()),
    #     bases=[model_cls],
    #     create_orm_model=True,
    # )

    ready_model_cls = type(
        f"{model_cls.__name__}Model",
        (_RowORMModelMixin, model_cls),
        {},
    )

    model_cls.__table_name__ = table_name  # type: ignore
    ready_model_cls.__table_name__ = table_name  # type: ignore
    ready_model_cls.__relation_fields__ = tuple(relation_names)  # type: ignore
    ready_model_cls.__column_fields__ = tuple(  # type: ignore
        table_columns.keys()
    )

    return table_data, ready_model_cls



"""Module contain DB component."""
from __future__ import annotations

import logging
import sqlite3
from collections import defaultdict
from functools import cached_property
from functools import singledispatchmethod
from pathlib import Path
from queue import Queue
from threading import Event
from threading import Thread
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Iterator
from typing import MutableMapping
from typing import Sequence
from typing import Type
from typing import TypeVar
from typing import cast
from typing import overload

from typing_extensions import TypeAlias
from typing_extensions import dataclass_transform

from .enumcls import ResultFetch
from .operations import CreateTable
from .operations import Query
from .orm.model import create_table_and_data_cls_row
from .sql import SQLBase
from .table.column import Column
from .table.table import Table


T = TypeVar("T")

if TYPE_CHECKING:
    from .orm.model import TableData
    TRegistredTables: TypeAlias = defaultdict[
        Any,
        list[tuple[TableData, type[Any]]]
    ]


__all__ = (
    "DB",
    "ThreadDB",
)


class DB:
    """DB component."""

    _instances: ClassVar[dict[Path, DB]] = {}
    _registered_tables_data: ClassVar[TRegistredTables] = defaultdict(
        list,
    )
    orm_classes: ClassVar[dict[str, type[Any]]] = {}

    def __new__(cls: type[DB], *args: Any, **kwargs: Any) -> DB:
        """Use singleton template. Check path and match paths."""
        if not args and kwargs.get("path") is None:
            msg = "DB.__init__() missing 1 required argument: 'path'"
            raise TypeError(msg)

        path = kwargs["path"] if kwargs.get("path") else args[0]
        normalized_path = cls.normalize_path(Path(path))

        # for inst_path, instance in cls._instances.items():
        #     if inst_path == normalized_path:
        #         return instance
        if normalized_path in cls._instances:
            return cls._instances[normalized_path]

        new_instance = super().__new__(cls)
        cls._instances[normalized_path] = new_instance
        return cls._instances[normalized_path]

    @classmethod
    def normalize_path(cls: type[DB], path: Path) -> Path:
        """Normalize path."""
        return path.parent.resolve().joinpath(path.name)

    def __init__(
        self,
        path: str,
        *,
        use_datacls: bool = True,
        debug: bool = False,
        auto_create_registred_tables: bool = True,
        custom_tables: list[Table] | None = None,
        **connect_params: Any,
    ) -> None:
        """Initialize DB create connection and cursor."""
        if debug:
            logging.basicConfig(level=logging.DEBUG)
        self.path = self.normalize_path(Path(path))
        self.connect: sqlite3.Connection = sqlite3.connect(
            self.path,
            **connect_params,
        )
        self.use_datacls = use_datacls
        self.table_names: set = set()
        self.create_table = getattr(self, "create_table", CreateTable)(self)
        self._custom_tables = custom_tables or []

        if auto_create_registred_tables:
            self.create_registred_table()

        self.initialize_tables()

    def initialize_tables(self) -> None:
        """Initialize all db tables."""
        stmt = (
            "SELECT name FROM sqlite_master "
            "WHERE type='table' AND substr(`name`, 1, 6) != 'sqlite';"
        )
        result = self.execute(stmt, result=ResultFetch.fetchall)

        custom_table_names = set()

        for table in self._custom_tables:
            custom_table_names.add(table.name.lower())
            setattr(
                self,
                table.name.lower(),
                table,
            )
            table(self)

        # table like class attributes
        # for attr in filter(
        #     lambda i: not i.startswith("_"),
        #     dir(self.__class__),
        # ):
        #     custom_table = getattr(self, attr)
        #     if not isinstance(custom_table, Table):
        #         continue
        #     custom_table_names.add(custom_table.name.lower())
        #     custom_table(self)

        # auto create exists table
        for name in result:
            table_name = name[0].lower()
            self.table_names.add(table_name)

            if table_name in custom_table_names:
                continue

            new_table: Table = Table(name[0], use_datacls=self.use_datacls)
            new_table(self)
            setattr(
                self,
                table_name,
                new_table,
            )
            self.table_names.add(table_name)

        if hasattr(self, "tables"):
            del self.tables

    @cached_property
    def tables(self) -> tuple[Table]:
        """Return all tables obj."""
        return tuple(
            getattr(self, table_name)
            for table_name in self.table_names
        )

    def __iter__(self) -> Iterator[Any]:
        """Iterate by db tables."""
        return self.tables.__iter__()

    def drop_tables(self) -> None:
        """Drop all db tables."""
        for table in self.tables:
            table.drop(init_tables=False)
        self.initialize_tables()

    @overload
    def execute(
        self,
        query: str,
        parameters: MutableMapping | Sequence = (),
        *,
        many: bool = False,
        size: int | None = None,
        result: ResultFetch | None = ResultFetch,
    ) -> list[Any]: ...

    @overload
    def execute(
        self,
        query: str,
        parameters: MutableMapping | Sequence = (),
        *,
        many: bool = False,
        size: int | None = None,
        result: ResultFetch | None = None,
    ) -> None: ...

    def execute(
        self,
        query: str,
        parameters: MutableMapping | Sequence = (),
        *,
        many: bool = False,
        size: int | None = None,
        result: ResultFetch | None = None,
    ) -> list[Any] | None:
        """Single execute to simplify it.

        Args:
            query (str): sql query
            parameters (MutableMapping | Sequence): data for executing.
            Defaults to ().
            many (bool): flag for executemany operation. Defaults to False.
            size (int | None): size for fetchmany operation. Defaults to None.
            result (ResultFetch | None): enum for fetch func. Defaults to None.

        Returns:
            list[Any] or None

        """
        command = query.partition(" ")[0].lower()
        cursor = self.connect.cursor()
        logging.info("Query: %s", query)
        logging.info("Parameters: %s", parameters)
        if many:
            cursor.executemany(query, parameters)
        else:
            cursor.execute(query, parameters)

        if (
            command in {"delete", "update", "create", "drop", "insert"} and
            result is None
        ):
            self.connect.commit()

        if command in {"drop", "create"}:
            self.initialize_tables()

        # Check result
        if result is None:
            cursor.close()
            return None

        ResultFetch(result)

        result_func: Callable = getattr(cursor, result.value)

        if result.value == "fetchmany":
            return result_func(size=size)

        data = result_func()

        cursor.close()
        if command == "insert":
            self.connect.commit()

        return data

    def close(self) -> None:
        """Close connection."""
        self.connect.close()

    def __enter__(self) -> DB:
        """Create context manager."""
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        """Close connection."""
        self.close()

    if TYPE_CHECKING:
        def __getattr__(self, name: str) -> Table:
            """Typing for runtime created table."""
            ...

    def create_registred_table(self) -> None:
        """Create registred table."""
        # cls = self.__class__
        cursor = self.connect.cursor()

        for path, tables in self._registered_tables_data.items():
            for table_info in tables:
                if path is None or path == self.path:

                    table_data, row_cls = table_info
                    table_name: str = table_data["table_name"]

                    create_table_query = self.create_table.query(**table_data)
                    logging.info("Query: %s", create_table_query)
                    cursor.execute(create_table_query)

                    table_obj = Table(
                        table_name,
                        use_datacls=True,
                        row_cls=row_cls,
                    )
                    # setattr(cls, table_name.lower(), table_obj)
                    self._custom_tables.append(table_obj)

        self.connect.commit()

    @dataclass_transform(kw_only_default=True)
    @overload
    def register_table(cls: Type[T]) -> Type[T]: ...

    @dataclass_transform(kw_only_default=True)
    @overload
    def register_table(
        *,
        path: str | None = None,
    ) -> Callable[[Type[T]], Type[T]]: ...

    @classmethod
    @dataclass_transform(kw_only_default=True)
    def register_table(
        cls,
        model_cls: Type[T] = None,
        /,
        *,
        path: str | Path | None = None,
    ) -> Type[T] | Callable[[Type[T]], Type[T]]:
        """Registrate table.

        Args:
            path (str | Path | None): path to datebase file
        """
        def wrap(model_cls: T) -> T:
            """Wrap func."""
            table_data_row_cls = create_table_and_data_cls_row(model_cls)
            correct_path = cls.normalize_path(Path(cast(str, path)))
            cls._registered_tables_data[correct_path].append(
                table_data_row_cls,
            )
            cls.orm_classes[table_data_row_cls[0]["table_name"]] = model_cls
            return cast(T, table_data_row_cls[1])

        if model_cls is None and path is not None:
            return wrap

        table_data_row_cls = create_table_and_data_cls_row(model_cls)
        cls._registered_tables_data[path].append(table_data_row_cls)
        cls.orm_classes[table_data_row_cls[0]["table_name"]] = model_cls
        return table_data_row_cls[1]

    def add(self, *orm_objs: Any) -> list[int | None]:
        """Add new ORM object in db."""
        if hasattr(orm_objs[0], "orm_obj") is False:
            msg = "Unknown object type"
            raise TypeError(msg)

        items_by_table: defaultdict[Table, list] = defaultdict(list)

        for obj in orm_objs:
            table_name: str = obj.__table_name__
            table_obj: Table = getattr(self, table_name)

            items_by_table[table_obj].append(obj)

        object_ids = []
        for table, objects in items_by_table.items():
            object_ids.append(table.add(*objects))

        return object_ids

    def query(self, *columns_or_orm: Column | Type[T] | SQLBase) -> Query[T]:
        """Create query."""
        query = None
        models = []
        columns = []

        if not columns_or_orm:
            msg = "Set columns or orm model"
            raise ValueError(msg)

        for col_orm in columns_or_orm:
            if isinstance(col_orm, Column):
                if query is None:
                    table: Table = getattr(
                        self,
                        col_orm.table_name.lower(),
                    )
                    query = Query(table)
                columns.append(col_orm)
                continue

            if isinstance(col_orm, SQLBase):
                if query is None:
                    table = getattr(
                        self,
                        col_orm.__table_name__,  # type: ignore
                    )
                    columns.append(col_orm)
                    query = Query(table)
                    continue

            if query is None:
                table = getattr(
                    self,
                    col_orm.__table_name__,  # type: ignore
                )
                query = Query(table)
            models.append(col_orm)

        if not columns and not models:
            msg = "Incorrect type"
            raise TypeError(msg)

        if columns and models:
            columns.extend(
                getattr(model, column)
                for model in models
                for column in model.__column_fields__  # type: ignore
            )
            return query(*columns)

        if columns:
            return query(*columns)

        if len(models) > 1:
            columns.extend(
                getattr(model, column)
                for model in models
                for column in model.__column_fields__  # type: ignore
            )
            return query(*columns)

        if query is None:
            msg = "Not found data for query"
            raise ValueError(msg)

        return query()


class Future:
    """Future for managing query execution."""

    def __init__(self) -> None:
        """Initialize."""
        self.event = Event()
        self.exception: Exception | None = None
        self.result: list[Any] | None = None

    def wait(self) -> None:
        """Wait for execution."""
        self.event.wait()

    @singledispatchmethod
    def put(self, result: list[Any] | None) -> None:
        """Write operation result."""
        self.result = result
        self.event.set()

    @put.register(Exception)
    def _(self, result: Exception) -> None:
        """Write exception."""
        self.exception = result
        self.event.set()

    def done(self) -> bool:
        """Check operation complited."""
        return self.event.is_set()


class ThreadDB(DB):
    """Thread safety db cls."""

    def __init__(
        self,
        path: str,
        *,
        use_datacls: bool = False,
        debug: bool = False,
        **connect_params: Any,
    ) -> None:
        """Initialize DB create connection, cursor and worker thread."""
        connect_params["check_same_thread"] = False
        self.worker_event = Event()
        self.worker_queue: Queue = Queue()
        self.worker = Thread(
            target=self.execute_worker,
            daemon=True,
        )
        self.worker.start()
        super().__init__(
            path,
            use_datacls=use_datacls,
            debug=debug,
            **connect_params,
        )
        if debug:
            logging.basicConfig(level=logging.DEBUG)

    def execute_worker(self) -> None:
        """Worker for executing sql-query and return result."""
        while not self.worker_event.is_set():
            try:
                future, args, kwargs = self.worker_queue.get()
                if kwargs.get("finish_worker"):
                    self.worker_event.set()
                    continue
                result = super().execute(*args, **kwargs)
                future.put(result)
            except Exception as e:  # noqa: PERF203
                logging.exception(
                    "Error: %s, Arguments: %s, %s",
                    e,
                    args,
                    kwargs,
                )
                future.put(e)
                self.connect.rollback()
            finally:
                future.event.set()
                self.worker_queue.task_done()

    def execute(  # type: ignore[override]
        self,
        *args: Any,
        **kwargs: Any,
    ) -> list[Any] | None:
        """Create future obj and sending args in worker."""
        future = Future()
        self.worker_queue.put((future, args, kwargs))
        future.wait()
        if future.done():
            return future.result
        return None

    def close(self) -> None:
        """Close worker thread and close db connection."""
        self.execute(finish_worker=True)
        self.worker.join()
        super().close()


if __name__ == "__main__":
    db = DB("local")



"""Module contains lib enum cls."""
from enum import Enum


__all__ = (
    "ResultFetch",
    "UpdateAction",
    "DeleteAction",
)


class ResultFetch(Enum):
    """Enum for fetching data from DB."""

    fetchmany = "fetchmany"
    fetchall = "fetchall"
    fetchone = "fetchone"


class UpdateAction(Enum):
    """On update action for foreign key column."""

    cascade = "cascade"
    set_null = "set null"
    restrict = "restrict"


class DeleteAction(Enum):
    """On delete action for foreign key column."""

    set_default = "set default"
    cascade = "cascade"
    set_null = "set null"
    restrict = "restrict"



"""Module contain orm classes."""
from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import Literal
from typing import Protocol
from typing import Sequence
from typing import TypeVar
from typing import cast
from typing import overload

from ..column_types import DataClassJson
from ..column_types import Date
from ..column_types import DateTime
from ..column_types import ForeignKey
from ..column_types import Json
from ..column_types import TColumnType
from ..column_types import Time
from ..table.column import Column
from .typings import TModel


if TYPE_CHECKING:
    from ..db import DB
    from ..rows import ABCRow
    from ..table.table import Table


T = TypeVar("T", bound=TModel)
TAnyType = TypeVar("TAnyType", covariant=True)


class TColumn(Protocol, Generic[TAnyType]):
    """Typed Column."""

    __slots__ = ()

    @overload
    def __get__(
        self,
        instance: None,
        owner: type[Any],
    ) -> Column:
        ...

    @overload
    def __get__(
        self,
        instance: Any,
        owner: type[Any],
    ) -> TAnyType:
        ...


class DBTableGetterMixin(Generic[T]):
    """Mixin to get db or table from current row instance."""

    __slots__ = ()

    def _get_db(self, instance: T) -> DB:
        """Get instance db or None."""
        if instance.table:
            return instance.table.db

        msg = f"{instance} is not linked to the database"
        raise AssertionError(msg)

    def _get_table_by_name(self, db: DB, table_name: str) -> Table:
        """Get table by table name."""
        table: Table | None = getattr(db, table_name.lower(), None)
        if table:
            return table
        msg = f"{table_name} not found"
        raise AssertionError(msg)


class MColumn(TColumn[TAnyType]):
    """Model column discriptor."""

    __slots__ = (
        "column_type",
        "_column_name",
        "_column",
        "_cache",
        "_is_cached",
        "_column_name_str",
    )

    _cached_types = {
        Json,
        DateTime,
        Date,
        Time,
        DataClassJson,
    }

    def __init__(self, column_type: TColumnType) -> None:
        self.column_type = column_type
        self._is_cached = column_type.__class__ in self._cached_types
        self._cache = None

    def __set_name__(self, owner: type[Any], column_name: str) -> None:
        self._column_name = f"_column_data_{column_name}_"
        self._column_name_str = column_name
        # table_name = owner.__name__
        # self._column = Column(table_name, column_name)

    def get_column(self, owner_cls: type[Any]) -> Column:
        """Get column object."""
        return Column(owner_cls.__table_name__, self._column_name_str)

    @overload
    def __get__(
        self,
        instance: None,
        owner: type[Any],
    ) -> Column:
        ...

    @overload
    def __get__(
        self,
        instance: Any,
        owner: type[Any],
    ) -> TAnyType:
        ...

    def __get__(
        self,
        instance: Any,
        owner: type[Any],
    ) -> TAnyType | Column:
        if instance is None:
            # return self._column
            return self.get_column(owner)

        # check cache
        if self._is_cached and self._cache:
            return self._cache

        value = getattr(instance, self._column_name)
        data = self.column_type.to_python(value)

        if self._is_cached is False:
            return data

        self._cache = data

        return data

    def __set__(self, instance: ABCRow, value: Any) -> None:
        """Set value."""
        if self._is_cached:
            self._cache = None
        setattr(
            instance,
            self._column_name,
            self.column_type.to_db(value),
        )


class RelationForeignKey(ForeignKey, DBTableGetterMixin, Generic[TAnyType]):
    """Relation foreign key."""

    __slots__ = ()

    def _add_new_foreign_object(
        self,
        instance: Any,
        new_foreign_obj: ABCRow,
    ) -> None:
        """Add new foreing object"""
        db = self._get_db(instance)
        table = self._get_table_by_name(db, self.second_table)
        new_obj = table.add(new_foreign_obj, returning=True)
        self.__set__(instance, new_obj)

    def __get__(
        self,
        instance: Any,
        owner: type[Any],
    ) -> TAnyType | None:
        """Returns a relation object by foreign key."""
        if instance is None:
            msg = "Instance not found"
            raise AttributeError(msg)

        db = self._get_db(instance)
        table = self._get_table_by_name(db, self.second_table)

        ref_column = getattr(table.c, self.reference_column.lower(), None)

        if ref_column is None:
            msg = f"Column '{self.reference_column}' not found"
            raise AttributeError(msg)

        foreign_key_value = getattr(instance, self.column, None)

        if foreign_key_value is None:
            return None

        return table.query().where(ref_column == foreign_key_value).first()

    def __set__(self, instance: ABCRow, value: ABCRow | None) -> None:
        """Set value."""
        if value is None:
            setattr(instance, self.column, value)
            return

        if hasattr(value, "orm_obj") is False:
            msg = "Incorrect object"
            raise TypeError(msg)

        if value.table is None:
            getattr(
                instance,
                "_relation_events",
            )["RelationForeignKey"].append(
                lambda: self._add_new_foreign_object(instance, value),
            )

        ref_table_value = getattr(value, self.reference_column)

        setattr(instance, self.column, ref_table_value)


class Relation(DBTableGetterMixin[T]):
    """
    Relation is a descriptor for dealing
    with one-to-many or many-to-many relationships.
    """

    __slots__ = (
        "second_table",
        "_foreign_key_to_current_table",
        "_foreign_key_to_relation_table",
        "_cascade",
    )

    def __init__(
        self,
        second_table: str,
        foreign_key_to_current_table: str,
        foreign_key_to_relation_table: str | None = None,
        cascade: Literal["delete", "update"] = "delete",
    ) -> None:
        """."""
        self.second_table = second_table.lower()
        self._foreign_key_to_current_table = foreign_key_to_current_table
        self._foreign_key_to_relation_table = foreign_key_to_relation_table
        self._cascade = cascade

    def _one_many(self, instance: T) -> list[T]:
        """One many."""
        db = self._get_db(instance)
        second_table = self._get_table_by_name(db, self.second_table)
        second_orm_model = db.orm_classes[self.second_table]

        second_tb_foreign_key: RelationForeignKey = second_orm_model.__dict__[
            self._foreign_key_to_current_table
        ]

        foreign_key_column: Column = getattr(
            second_orm_model,
            second_tb_foreign_key.column,
        )

        instance_ref_column_value = getattr(
            instance,
            second_tb_foreign_key.reference_column,
        )

        return second_table.query().where(
            foreign_key_column == instance_ref_column_value,
        ).all()

    def _many_to_many(self, instance: T) -> list[T]:
        """Many to many."""
        db = self._get_db(instance)

        m2m_table = self._get_table_by_name(db, self.second_table)
        m2m_orm_model = db.orm_classes[self.second_table]

        current_tb_foreign_key: RelationForeignKey = m2m_orm_model.__dict__[
            self._foreign_key_to_current_table
        ]

        relation_tb_foreign_key: RelationForeignKey = m2m_orm_model.__dict__[
            cast(str, self._foreign_key_to_relation_table)
        ]

        relation_table = self._get_table_by_name(
            db,
            relation_tb_foreign_key.second_table,
        )
        rel_orm_model = db.orm_classes[relation_tb_foreign_key.second_table]

        instance_foreing_key_value = getattr(
            instance,
            current_tb_foreign_key.reference_column,
        )

        rel_column: Column = getattr(
            m2m_orm_model,
            relation_tb_foreign_key.column,
        )
        cur_column: Column = getattr(
            m2m_orm_model,
            current_tb_foreign_key.column,
        )

        # TODO(stickking): join??
        # 0000
        query_m2m = m2m_table.query(
            rel_column,
        ).where(
            cur_column == instance_foreing_key_value,
        ).all(
            only_data=True,
        )

        relation_ids = [id_[0] for id_ in query_m2m]

        if not relation_ids:
            return []

        rel_column = getattr(
            rel_orm_model,
            relation_tb_foreign_key.reference_column,
        )

        return relation_table.query().where(rel_column.in_(relation_ids)).all()

    def _prepare_relation_objects(
        self,
        relation_table: Table,
        objects: Sequence[T],
    ) -> list[T]:
        """Prepare relation objects."""
        exists_object = [obj for obj in objects if obj.table is not None]
        new_objects = [obj for obj in objects if obj.table is None]

        added_objects = [
            cast(Any, relation_table.add(new_obj, returning=True))
            for new_obj in new_objects
        ]

        return [*exists_object, *added_objects]

    def _add_new_one_many_object(
        self,
        instance: Any,
        objects: Sequence[T],
    ) -> None:
        """Add new relation objects."""
        db = self._get_db(instance)
        table = self._get_table_by_name(db, self.second_table)
        second_orm_model = db.orm_classes[self.second_table]
        second_tb_foreign_key: RelationForeignKey = second_orm_model.__dict__[
            self._foreign_key_to_current_table
        ]

        ready_objects = self._prepare_relation_objects(table, objects)

        ref_value = getattr(instance, second_tb_foreign_key.reference_column)

        id_values = [
            {name: getattr(obj, name) for name in table.primary_keys}
            for obj in ready_objects
        ]

        condition = " OR ".join(
            table.update._make_operator_query(pr_key, without_parameters=True)
            for pr_key in id_values
        )

        if self._cascade == "update":
            table.update(
                {second_tb_foreign_key.column: None},
                **{second_tb_foreign_key.column: ref_value},
            )
        else:
            table.delete(**{second_tb_foreign_key.column: ref_value})

        table.update(
            {second_tb_foreign_key.column: ref_value},
            condition=condition,
        )

    def _add_new_many_to_many_object(
        self,
        instance: Any,
        objects: Sequence[T],
    ) -> None:
        """Add new relation objects."""
        db = self._get_db(instance)

        m2m_table = self._get_table_by_name(db, self.second_table)
        m2m_orm_model = db.orm_classes[self.second_table]

        current_tb_foreign_key: RelationForeignKey = m2m_orm_model.__dict__[
            self._foreign_key_to_current_table
        ]
        relation_tb_foreign_key: RelationForeignKey = m2m_orm_model.__dict__[
            cast(str, self._foreign_key_to_relation_table)
        ]
        relation_table = self._get_table_by_name(
            db,
            relation_tb_foreign_key.second_table,
        )

        ready_objects = self._prepare_relation_objects(relation_table, objects)

        instance_ref_column_value = getattr(
            instance,
            current_tb_foreign_key.reference_column,
        )

        m2m_current_fk_column = current_tb_foreign_key.column
        m2m_relation_fk_column = relation_tb_foreign_key.column

        relation_ref_col_name = relation_tb_foreign_key.reference_column

        m2m_table.delete(**{
            m2m_current_fk_column: instance_ref_column_value,
        })

        m2m_table.add(*[
            {
                m2m_current_fk_column: instance_ref_column_value,
                m2m_relation_fk_column: getattr(obj, relation_ref_col_name)
            }
            for obj in ready_objects
        ])

    def __get__(
        self,
        instance: T | None,
        owner: type[Any],
    ) -> list[T]:
        """Returns a relation objects by foreign keys."""
        if instance is None:
            msg = "Instance not found"
            raise AttributeError(msg)

        if self._foreign_key_to_relation_table is not None:
            return self._many_to_many(instance)

        return self._one_many(instance)

    def __set__(
        self,
        instance: T,
        value: Sequence[T],
    ) -> None:
        """Set new relation objects."""
        if self._foreign_key_to_relation_table is not None:
            getattr(
                instance,
                "_relation_events",
            )["Relation"].append(
                lambda: self._add_new_many_to_many_object(instance, value),
            )
            return

        getattr(
            instance,
            "_relation_events",
        )["Relation"].append(
            lambda: self._add_new_one_many_object(instance, value),
        )



"""Module contains base operation classes."""
from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from functools import wraps
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Generator
from typing import Generic
from typing import Iterable
from typing import Literal
from typing import MutableMapping
from typing import Sequence
from typing import TypeVar
from typing import overload

from .column_types import BaseType
from .enumcls import ResultFetch
from .rows import create_result_row


if TYPE_CHECKING:
    from .column_types import ForeignKey
    from .db import DB
    from .sql import SQLBase
    from .table import Column
    from .table import Table

    TOperator = Literal["AND", "and", "OR", "or", ","]
    TQueryData = dict[str, int | bool | str | None]
    F = TypeVar("F", bound=Callable)


__all__ = (
    "Query",
    "Select",
    "Insert",
    "Delete",
    "Update",
    "CreateTable",
)


TRow = TypeVar("TRow")


class Operation(ABC):
    """Base operation."""

    __slots__ = ()

    @abstractmethod
    def __init__(self) -> None:
        ...

    def _make_operator_query(
        self,
        data: TQueryData,
        operator: TOperator = "AND",
        *,
        without_parameters: bool = False,
    ) -> str:
        """Build an sql expression with 'and' and 'or' operators."""
        if operator.lower() not in {"and", "or", ","}:
            msg = "Incorrect operator."
            raise ValueError(msg)

        if not without_parameters:
            return f" {operator} ".join(
                f"{key} is NULL" if value is None else f"{key} = :{key}"
                for key, value in data.items()
            )

        return f" {operator} ".join(
            f"{key} is NULL"
            if value is None else
            f"{key} = '{value}'"
            if isinstance(value, str)
            else f"{key} = {value}"
            for key, value in data.items()
        )

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...


class TableOperation(Operation, ABC):
    """Base operation."""

    __slots__ = ("table",)

    def __init__(self, table: Table) -> None:
        self.table = table

    def _make_operator_query(
        self,
        data: TQueryData,
        operator: TOperator = "AND",
        *,
        without_parameters: bool = False,
        is_null: bool = True,
    ) -> str:
        if operator.lower() not in {"and", "or", ","}:
            msg = "Incorrect operator."
            raise ValueError(msg)

        if not without_parameters:
            return f" {operator} ".join(
                f"{key} is NULL" if (
                    value is None and is_null
                ) else f"{key} = :{key}"
                for key, value in data.items()
            )

        return f" {operator} ".join(
            f"{key} is NULL"
            if (
                value is None and is_null
            ) else
            f"{key} = '{value}'"
            if isinstance(value, str)
            else f"{key} = {value}"
            for key, value in data.items()
        )

    @abstractmethod
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        ...


class Select(TableOperation, Generic[TRow]):
    """Component for select and filtered DB data."""

    __slots__ = ()

    def query(self, columns: Sequence[str] | None = None) -> str:
        """Fetch base query."""
        columns_str = self._generate_columns(columns)
        return f"SELECT {columns_str} FROM {self.table.name}"  # noqa: S608

    def _generate_columns(
        self,
        columns: Sequence[str] | None = None,
        table: Table | None = None,
    ) -> str:
        """Create column str."""
        columns_names: Sequence[str] = self.table.column_names
        table_name = self.table.name
        if columns:
            columns_names = columns
        return ", ".join(
            f"`{table_name}`.{name}"
            for name in columns_names
        )

    def _execute(
        self,
        query: str,
        parameters: TQueryData,
        *,
        size: int = 0,
        columns: Sequence[str] | None = None,
        return_generator: bool = False,
    ) -> list[TRow]:
        """Execute with size."""
        # result = None
        if size:
            result = self.table.execute(
                query,
                parameters,
                size=size,
                result=ResultFetch.fetchmany,
            )
        else:
            result = self.table.execute(
                query,
                parameters,
                result=ResultFetch.fetchall,
            )
        if return_generator:
            self._as_generator_row(result, columns=columns)
        return self._as_list_row(result, columns=columns)

    def _as_list_row(
        self,
        items: Sequence[tuple[tuple[Any, ...]]],
        *,
        columns: Sequence[str] | None = None,
    ) -> list[TRow]:
        """Create list rows."""
        row_cls = self.table.row_cls
        columns_name: Sequence[str] = self.table.column_names
        if columns:
            row_cls = create_result_row(columns)
            columns_name = columns
        return [
            row_cls(
                table=self.table,
                **dict(zip(columns_name, item)),
            )
            for item in items
        ]

    def _as_generator_row(
        self,
        items: Sequence[tuple[tuple[Any, ...]]],
        *,
        columns: Sequence[str] | None = None,
    ) -> Generator[TRow, Any, None]:
        """Create rows generator."""
        row_cls = self.table.row_cls
        columns_name: Sequence[str] = self.table.column_names
        if columns:
            row_cls = create_result_row(columns)
            columns_name = columns
        for item in items:
            yield row_cls(
                table=self.table,
                **dict(zip(columns_name, item)),
            )

    def _filter(
        self,
        filter_by: TQueryData,
        *,
        operator: TOperator = "AND",
        columns: Sequence[str] | None = None,
    ) -> str:
        """Filter data by filters value where key is
        column name value is content.
        """
        operator_query = self._make_operator_query(
            filter_by,
            operator,
        )
        return f"{self.query(columns)} WHERE {operator_query}"

    def __call__(
        self,
        *,
        size: int = 0,
        operator: TOperator = "AND",
        columns: Sequence[str] | None = None,
        condition: str | None = None,
        return_generator: bool = False,
        **filter_by: int | str | None,
    ) -> list[TRow]:
        """Select-query for current table."""
        query = self.query(columns)
        if filter_by:
            query = self._filter(
                filter_by,
                operator=operator,
                columns=columns,
            )
        if condition:
            query = "{} WHERE {}".format(
                self.query(columns),
                condition,
            )
        return self._execute(
            query,
            filter_by,
            size=size,
            columns=columns,
            return_generator=return_generator,
        )


class Insert(TableOperation, Generic[TRow]):
    """Component for insert data in DB."""

    __slots__ = ()

    def _as_list_row(
        self,
        items: Sequence[tuple[tuple[Any, ...]]],
    ) -> list[TRow]:
        """Create list rows."""
        row_cls = self.table.row_cls
        columns_name: Sequence[str] = self.table.column_names
        return [
            row_cls(
                table=self.table,
                **dict(zip(columns_name, item)),
            )
            for item in items
        ]

    def query(
        self,
        data: Sequence[TQueryData],
        *,
        returning: bool = False,
    ) -> str:
        """Create insert sql-query."""
        query = ", ".join(
            f":{key}"
            for key in data[0]
        )
        colums_name = ", ".join(
            name
            for name in data[0]
        )
        returning_str = ""
        if returning:
            all_column_names = ", ".join(
                name
                for name in self.table.column_names
            )
            returning_str = "RETURNING " + all_column_names
        return "INSERT INTO `{}` ({}) VALUES({}) {}".format(
            self.table.name,
            colums_name,
            query,
            returning_str,
        )

    @overload
    def __call__(
        self,
        data: TQueryData | Sequence[TQueryData],
        *,
        returning: Literal[False] = False,
    ) -> None:
        ...

    @overload
    def __call__(
        self,
        data: TQueryData | Sequence[TQueryData],
        *,
        returning: Literal[True] = True,
    ) -> TRow:
        ...

    def __call__(
        self,
        data: TQueryData | Sequence[TQueryData],
        *,
        returning: bool = False,
    ) -> TRow | None:
        """Insert-query for current table."""
        if not data:
            msg = "Data do not be empty."
            raise ValueError(msg)

        if isinstance(data, dict):
            data = (data,)

        if returning is False:
            self.table.execute(self.query(data), data, many=True)
            return None

        if len(data) > 1:
            msg = "Returning work with one item"
            raise ValueError(msg)

        result_items = self.table.execute(
            self.query(data, returning=returning),
            data[0],
            result=ResultFetch.fetchall,
        )
        items = self._as_list_row(result_items)
        return items[0]


class Delete(TableOperation):
    """Component for delete row from db."""

    __slots__ = ()

    def query(self) -> str:
        """Fetch base delete query."""
        return f"DELETE FROM `{self.table.name}` WHERE id=?"  # noqa: S608

    def _filter(
        self,
        filter_by: TQueryData,
        *,
        operator: TOperator = "AND",
    ) -> None:
        """Filter delete row from table."""
        if not filter_by:
            msg = "Value do not be empty."
            raise ValueError(msg)
        query_and = self._make_operator_query(filter_by, operator)
        query = "DELETE FROM `{}` WHERE {}".format(
            self.table.name,
            query_and
        )
        self.table.execute(query, filter_by)

    def __call__(
        self,
        id: int | Sequence[int] | None = None,  # noqa: A002
        *,
        operator: TOperator = "AND",
        condition: str | None = None,
        **filter_by: int | str | None,
    ) -> None:
        """Delete-query for current table."""
        if isinstance(id, Sequence):
            ids = tuple((id_,) for id_ in id)
            self.table.execute(self.query(), ids, many=True)  # type: ignore
            return
        if id is not None:
            filter_by["id"] = id

        if condition:
            query = f"DELETE FROM `{self.table.name}` WHERE {condition}"
            self.table.execute(query)  # type: ignore
            return

        self._filter(filter_by, operator=operator)


class Update(TableOperation):
    """Component for updating table row."""

    __slots__ = ("query",)

    def __init__(self, table: Table) -> None:
        super().__init__(table)
        self.query = f"UPDATE `{self.table.name}` SET "

    def __call__(
        self,
        data: TQueryData,
        operator: TOperator = "AND",
        condition: str | None = None,
        **filter_by: Any,
    ) -> None:
        """Insert-query for current table."""
        print(data)
        if not isinstance(data, dict):
            msg = "Incorrect type for 'data.'"
            raise TypeError(msg)
        if not data:
            msg = "Argument 'data' do not be empty."
            raise ValueError(msg)
        new_column_values = self._make_operator_query(
            data,
            operator=",",
            is_null=False,
        )
        filter_value = self._make_operator_query(
            filter_by,
            operator,
            without_parameters=True,
        )
        query = self.query + new_column_values
        if filter_by:
            query = f"{query} WHERE {filter_value}"
            self.table.execute(query, data)  # type: ignore
            return
        if condition:
            query = f"{query} WHERE {condition}"
        self.table.execute(query, data)  # type: ignore


def generate_query(
    func: F,
) -> F:
    """Generate new query object."""
    moved_attr_names = (
        "_body",
        "_filters",
        "_having",
        "_orders",
        "_groups",
        "_limit",
        "_offset",
        "columns",
    )

    @wraps(func)
    def wrap(self: Query, *args: Any, **kwargs: Any) -> Query:
        """Wrap func."""
        new_query = Query(self.table)
        for attr_name in moved_attr_names:
            setattr(
                new_query,
                attr_name,
                getattr(self, attr_name)
            )
        new_query_func = getattr(new_query, func.__name__).__wrapped__
        return new_query_func(new_query, *args, **kwargs)

    return wrap  # type: ignore


class Query(TableOperation, Generic[TRow]):
    """Create sql query with more params."""

    __slots__ = (
        "_body",
        "_filters",
        "_having",
        "_orders",
        "_groups",
        "_limit",
        "_offset",
        "columns",
    )

    _body: str
    _filters: tuple[Any, ...]
    _having: tuple[Any, ...]
    _orders: tuple[Any, ...]
    _groups: tuple[Any, ...]
    _limit: int
    _offset: int
    columns: tuple[SQLBase | Column, ...]

    def __init__(
        self,
        table: Table[TRow] | None = None,
    ) -> None:
        """Initialize query and create it body

        Args:
            rowcls (ABCRow | None, optional): row cls for existents table.
                Defaults to None.
        """
        self.table = table  # type: ignore

    def _create_query_str(self) -> str:
        """Create sql query with all existent attrs."""
        where_str = " ".join(map(lambda i: str(i), self._filters))
        if where_str:
            where_str = " WHERE " + where_str

        group_by_str = ""
        if self._groups:
            group_by_str = " GROUP BY " + ", ".join(self._groups)

        limit_str = ""
        if self._limit:
            limit_str = f" LIMIT {self._limit}"
            if self._offset:
                limit_str += f" OFFSET {self._offset}"

        having_str = " ".join(self._having)
        if having_str:
            having_str = " HAVING " + having_str

        order_by_str = ", ".join(self._orders)
        if order_by_str:
            order_by_str = " ORDER BY " + order_by_str

        table = self.table.name
        return "SELECT {} FROM `{}`{}{}{}{}{}".format(
            self._body,
            table,
            where_str,
            group_by_str,
            having_str,
            order_by_str,
            limit_str,
        )

    __str__ = _create_query_str

    def _as_list_row(
        self,
        items: Sequence[tuple[tuple[Any, ...]]],
    ) -> list[TRow]:
        """Create list rows."""
        row_cls = self.table.row_cls
        columns_name: Sequence[str] = self.table.column_names
        if self.columns:
            columns_name = self.result_row_column_names
            row_cls = create_result_row(columns_name)
        return [
            row_cls(
                table=self.table,
                **dict(zip(columns_name, item)),
            )
            for item in items
        ]

    @property
    def result_row_column_names(self) -> list[str]:
        """Prepare column names for class row."""
        return [
            col.row_name
            for col in self.columns
        ]

    def _prepare_column(self, column: str | SQLBase | Column) -> str:
        """Prepare one column."""
        # if column is Column or SQLBase
        column = str(column)
        if column in self.table.column_names:
            return f"`{self.table.name}`.{column}"
        return column

    def _generate_column_names(
        self,
        columns: tuple[SQLBase | Column, ...] | None = None,
    ) -> str:
        """Create column str."""
        columns_names = self.table.column_names  # type: ignore
        # table_name = self.table.name  # type: ignore
        if columns:
            columns_names = columns  # type: ignore
        return ", ".join(
            self._prepare_column(name)
            for name in columns_names
        )

    def _execute(self, query: str, size: int | None = None) -> list[tuple]:
        """Execute query."""
        if size:
            return self.table.execute(
                query,
                size=size,
                result=ResultFetch.fetchmany,
            )
        return self.table.execute(
            query,
            result=ResultFetch.fetchall,
        )

    @generate_query
    def limit(self, limit_number: int) -> Query[TRow]:
        """Use limit in sql query."""
        if not isinstance(limit_number, int):
            msg = f"Limit not be {type(limit_number)}"
            raise TypeError(msg)
        self._limit = limit_number
        return self

    @generate_query
    def offset(self, offset_number: int) -> Query[TRow]:
        """Use offset in sql query."""
        if not isinstance(offset_number, int):
            msg = f"Offset not be {type(offset_number)}"
            raise TypeError(msg)
        self._offset = offset_number
        return self

    @generate_query
    def order_by(
        self,
        *args: str,
        **orders: Literal["asc", "desc"],
    ) -> Query[TRow]:
        """Use order by in query."""
        order_types = {"asc", "desc"}
        if args:
            self._orders += tuple(map(lambda i: str(i), args))
            return self

        assert all(value.lower() in order_types for value in orders.values())
        self._orders += tuple(
            f"{key} {value}"
            for key, value in orders.items()
        )
        return self

    def _sum_tuples(self, attr_name: str, value: tuple) -> None:
        """Summing tuples."""
        setattr(
            self,
            attr_name,
            getattr(self, attr_name) + value,
        )

    def _add_condition(
        self,
        condition: str,
        operator: TOperator = "AND",
        in_having: bool = False,
    ) -> None:
        """Adding condition in _filter."""
        attr_name = "_having" if in_having else "_filters"
        condition_orig = condition.strip()
        condition = condition.strip().lower()
        # if self._filters:
        if getattr(self, attr_name):
            if (
                condition.startswith("and") or
                condition.startswith("or")
            ):
                self._sum_tuples(attr_name, (condition_orig,))
                return
            condition_orig = f"{operator} {condition_orig}"
            self._sum_tuples(attr_name, (condition_orig,))
            return

        if condition.startswith("and"):
            condition_orig = condition_orig[3:].strip()
        elif condition.startswith("or"):
            condition_orig = condition_orig[1:].strip()
        self._sum_tuples(attr_name, (condition_orig,))

    def _add_filters(
        self,
        filter_by: Any,
        filter_operator: TOperator = "AND",
        operator: TOperator = "AND",
        in_having: bool = False,
    ) -> None:
        attr_name = "_having" if in_having else "_filters"
        filter_str = self._make_operator_query(
            filter_by,
            filter_operator,
            without_parameters=True,
        )
        # if self._filters:
        if getattr(self, attr_name):
            self._sum_tuples(attr_name, (f"{operator} {filter_str}",))
            # self._filters += (f"{operator} {filter_str}",)
            return
        self._sum_tuples(attr_name, (filter_str,))
        # self._filters += (filter_str,)

    @generate_query
    def where(
        self,
        *args: str,
        condition: str | None = None,
        filter_operator: TOperator = "AND",
        operator: TOperator = "AND",
        **filter_by: Any,
    ) -> Query[TRow]:
        """Use where construction in sql query."""
        # TODO (stickking): check operator and remove from it ','
        # 0000
        if args:
            for arg in args:
                self._add_condition(str(arg))
            return self
        if condition:
            self._add_condition(condition, operator)
            return self
        self._add_filters(filter_by, operator, filter_operator)
        return self

    @generate_query
    def having(
        self,
        *args: str,
        condition: str | None = None,
        filter_operator: TOperator = "AND",
        operator: TOperator = "AND",
        **filter_by: Any,
    ) -> Query[TRow]:
        """Use where construction in sql query."""
        # TODO (stickking): check operator and remove from it ','
        # 0000
        if args:
            for arg in args:
                self._add_condition(arg)
            return self
        if condition:
            self._add_condition(condition, operator, in_having=True)
            return self
        self._add_filters(filter_by, operator, filter_operator, in_having=True)
        return self

    @generate_query
    def group_by(self, *args: str | Column) -> Query[TRow]:
        """Use group by operation."""
        self._groups += tuple(map(str, args))
        return self

    def exists(self) -> bool:
        """Contain all query in exists command."""
        query = self._create_query_str()
        query = f"SELECT EXISTS({query})"
        result = self._execute(query)
        if not result or not result[0]:
            return False
        return bool(result[0][0])

    def count(self) -> int:
        """Contain all query in exists command."""
        first_column = self._body.split(", ")[0]
        if not first_column:
            return 0
        self._body = f"COUNT({first_column})"
        query = self._create_query_str()
        result = self._execute(query)
        if not result or not result[0]:
            return 0
        return result[0][0]

    def first(self) -> TRow | None:
        """Return first item from query."""
        self.limit(1)
        query = self._create_query_str()
        result = self._execute(query, 0)
        items = self._as_list_row(result)
        if not items:
            return None
        return items[0]

    one = first

    @overload
    def all(
        self,
        size: int = 0,
        *,
        only_data: Literal[False] = False,
    ) -> list[TRow]: ...

    @overload
    def all(
        self,
        size: int = 0,
        *,
        only_data: Literal[True] = True,
    ) -> list[tuple[Any, ...]]: ...

    def all(
        self,
        size: int = 0,
        *,
        only_data: bool = False,
    ) -> list[TRow] | list[tuple[Any, ...]]:
        """Return first item from query."""
        query = self._create_query_str()
        result = self._execute(query, size)
        if only_data:
            return result
        return self._as_list_row(result)

    def generative_all(self, limit: int) -> Generator[TRow, Any, None]:
        """Return first item from query."""
        row_cls = self.table.row_cls
        columns_name: Sequence[str] = self.table.column_names
        if self.columns:
            columns_name = self.result_row_column_names
            row_cls = create_result_row(columns_name)

        offset = 0
        query = self.limit(limit)
        while True:
            query2 = query.offset(offset * limit)
            items = self._execute(query2._create_query_str(), 0)
            offset += 1

            if not items:
                break

            for item in items:
                yield row_cls(
                    table=self.table,
                    **dict(zip(columns_name, item)),
                )

    def __iter__(self) -> Iterable[TRow]:
        """Iteration by data."""
        return iter(self.all())

    def __call__(
        self,
        *columns: SQLBase | Column,
        table: Table | None = None,
    ) -> Query[TRow]:
        """Initialize base args and create query body."""
        self._body = ""
        self._filters = ()
        self._having = ()
        self._orders = ()
        self._groups = ()
        self._limit = 0
        self._offset = 0
        self.columns = ()

        if self.table is None and table:
            self.table = table
            self._body = self._generate_column_names()
            return self

        if columns:
            self.columns = columns
            self._body = self._generate_column_names(columns)
            return self

        self._body = self._generate_column_names()
        return self


class CreateTable(Operation):
    """Create table object."""

    __slots__ = ("db", )

    def __init__(self, db: DB) -> None:
        """Initialize."""
        self.db = db

    def query(
        self,
        table_name: str,
        columns: Sequence[str] | MutableMapping[str, Any],
        table_primary_key: Sequence[str] | None = None,
        foreign_keys: Sequence[ForeignKey] | None = None,
        *,
        if_not_exists: bool = True,
    ) -> str:
        """Return SQL command."""
        columns = self._generate_columns(
            columns,
        )
        table_primary_keys = self._genarate_table_primary_keys(
            table_primary_key,
        )
        foreign_keys_str = self._genatate_table_foreign_keys(
            foreign_keys
        )

        query = "CREATE TABLE "
        if if_not_exists:
            query += "IF NOT EXISTS "

        # if table_primary_key and foreign_keys:
        #     foreign_keys = ", " + foreign_keys

        pr_fr_keys = table_primary_keys + foreign_keys_str

        return f"{query} `{table_name}` ({columns}{pr_fr_keys})"

    def _genarate_table_primary_keys(
        self,
        table_primary_key: Sequence[str] | None = None,
    ) -> str:
        """Create table primary keys."""
        if isinstance(table_primary_key, Sequence):
            return ", PRIMARY KEY(" + ",".join(
                _ for _ in table_primary_key
            ) + ")"
        return ""

    def _genatate_table_foreign_keys(
        self,
        foreign_keys: Sequence[ForeignKey] | None = None,
    ) -> str:
        """Create table foreign keys."""
        if foreign_keys is not None and isinstance(foreign_keys, Sequence):
            return ", " + ", ".join(
                key()
                for key in foreign_keys
            )
        return ""

    def _generate_columns(
        self,
        columns: Sequence[str] | MutableMapping[str, Any],
    ) -> str:
        """Create column name with type if it exists."""
        if (
            isinstance(columns, Sequence) and
            all(isinstance(_, str) for _ in columns)
        ):
            return ", ".join(columns)

        if (
            not isinstance(columns, MutableMapping) or
            not all(isinstance(_, BaseType) for _ in columns.values())
        ):
            msg = "Incorrect type for column item"
            raise TypeError(msg)

        return ", ".join(
            f"`{key}` {value}"
            for key, value in columns.items()
        )

    def __call__(
        self,
        table_name: str,
        columns: Sequence[str] | MutableMapping[str, Any],
        table_primary_key: Sequence[str] | None = None,
        foreign_keys: Sequence[ForeignKey] | None = None,
        *,
        if_not_exists: bool = True,
    ) -> None:
        """Create table in DB.

        Args:
            table_name (str): table name
            columns (Sequence[str] | MutableMapping[str, str]): column name or
            dict column with column types
            table_primary_key (Sequence[str] | None): set table primary key.
            Defaults to None.
            if_not_exists (bool): use 'if not exists' in query.
            Defaults to True.

        Raises:
            TypeError: Incorrect type for columns
            TypeError: Incorrect type for column item

        """
        if not isinstance(columns, (Sequence, MutableMapping)):
            msg = "Incorrect type for columns"
            raise TypeError(msg)

        query = self.query(
            table_name,
            columns,
            table_primary_key,
            foreign_keys,
            if_not_exists=if_not_exists,
        )
        self.db.execute(query)



"""Module contain lildb."""
from .db import DB  # noqa: F401
from .db import ThreadDB  # noqa: F401
from .enumcls import *  # noqa: F403
from .operations import *  # noqa: F403
from .rows import *  # noqa: F403
from .table import Column  # noqa: F401
from .table import Table  # noqa: F401



"""Module contain SQL functions and other operation."""
from __future__ import annotations

from typing import Any
from typing import Literal
from typing import Sequence
from typing import TypeVar


__all__ = (
    "func",
)


TFuncObj = TypeVar("TFuncObj", bound="SQLBase")


TFunc = Literal[
    "abs",
    "avg",
    "sum",
    "min",
    "max",
    "count",
    "random",
    "lower",
    "upper",
    "length",
    "distinct",
    "substr",
    "like",
    "ltrim",
    "rtrim",
    "trim",
    "replace",
    "instr",
    "date",
    "time",
    "datetime",
    "julianday",
    "strftime",
]


FUNC_NAMES = {
    "abs",
    "avg",
    "sum",
    "min",
    "max",
    "count",
    "random",
    "lower",
    "upper",
    "length",
    "distinct",
    "substr",
    "like",
    "ltrim",
    "rtrim",
    "trim",
    "replace",
    "instr",
    "date",
    "time",
    "datetime",
    "julianday",
    "strftime",
}


class SQLBase:
    """Base sql function or operation."""

    __slots__ = (
        "_data",
        "_label",
        "_template",
        "_disable_arguments",
        "_name",
        "__table_name__",
    )

    # template = "{func}({data}) AS {label}"

    def __init__(
        self,
        name: str,
        template: str | None = None,
        data: Any = None,
        *,
        disable_arguments: bool = False,
    ) -> None:
        """Initialize"""
        self._name = name
        self._template = template or "{func}({data}) AS {label}"
        self._data = data
        self._disable_arguments = disable_arguments
        self._label: str | None = None

    def __call__(self, *args: Any) -> SQLBase:
        """Call sql."""
        if self._disable_arguments and args:
            msg = "Function takes 0 positional arguments"
            raise TypeError(msg)

        self._data = args

        if self._data:
            self.__table_name__ = self._data[0].table_name
        return self

    def label(self, name: str) -> SQLBase:
        """Create AS label."""
        self._label = name
        return self

    @property
    def complete_label(self) -> str:
        """Prepare label for command"""
        if self._label:
            return self._label

        label = "_".join(
            column._column_name
            for column in self._data
            if hasattr(column, "_column_name")
        )

        if label:
            return label

        # label = self.__class__.__name__.lower()
        label = self._name.lower()
        return label

    @property
    def data(self) -> str | list[str]:
        """Return completed."""
        if "]}" in self._template:
            return [
                f"'{arg}'" if isinstance(arg, str) else str(arg)
                for arg in self._data
            ]
        return ", ".join(
            f"'{arg}'" if isinstance(arg, str) else str(arg)
            for arg in self._data
        )

    @data.setter
    def data(self, value: Sequence[Any]) -> None:
        """Return completed."""
        self._data = value

    def __str__(self) -> str:
        """Create string view"""
        # operation = self.__class__.__name__.upper()
        operation = self._name
        label = self.complete_label
        return self._template.format(**{
            "func": operation,
            "data": self.data,
            "label": label,
        })

    def __eq__(self, value: Any) -> bool:
        """Eq operation."""
        return str(self) == value

    @property
    def row_name(self) -> str:
        """Row name."""
        return self.complete_label


class Func:
    """Object to generation all sql funcs."""

    __slots__ = ()

    def __getattr__(self, name: TFunc | str) -> SQLBase:
        """Create func for column."""
        if name not in FUNC_NAMES:
            msg = "{} {} {}".format(
                self.__class__.__name__,
                "object has no attribute",
                name,
            )
            raise AttributeError(msg)
        # func_cls: type[TFuncObj] = type(
        #     name.upper(),
        #     (SQLBase,),
        #     {},
        # )

        name_lower = name.lower()

        if name_lower in "distinct":
            # func_cls.template = "{func} {data}"
            return SQLBase(
                name.upper(),
                template="{func} {data}",
            )

        if name_lower == "like":
            # func_cls.template = "{data[0]} {func} {data[1]}"
            return SQLBase(
                name.upper(),
                template="{data[0]} {func} {data[1]}",
            )

        if name_lower == "random":
            # def __init__(self) -> None:
            #     self._data = ""
            #     self._label = None
            # func_cls.__init__ = __init__
            return SQLBase(
                name.upper(),
                data="",
                disable_arguments=True,
            )
        return SQLBase(
            name.upper(),
        )


func = Func()



"""Module contains any funcs."""
from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import TypeVar


if TYPE_CHECKING:
    from ..table import Table
    from .typings import TModel
    TModelClass = TypeVar("TModelClass", bound=type[Any])


def contain_relation_objects(orm_object: TModel) -> bool:
    """Check object contain new relation objects."""
    return len(orm_object._relation_events) > 0


def refresh_old_obj_by_new(
    table: Table,
    old_orm_object: Any,
    new_orm_object: Any,
) -> None:
    """Move relation funcs to other object."""
    for name in table.column_names:
        col_name = f"_column_data_{name}_"
        setattr(
            old_orm_object,
            col_name,
            getattr(new_orm_object, col_name)
        )


def process_add_relation_objects(
    orm_object: Any,
    ref_type: Literal["Relation", "RelationForeignKey"] | None = None,
) -> None:
    """Add relation object."""
    if ref_type is None:
        for funcs in orm_object._relation_events.values():
            for func in funcs:
                func()
        orm_object._relation_events = defaultdict(list)
        return

    for func in orm_object._relation_events[ref_type]:
        func()

    orm_object._relation_events.pop(ref_type)



"""Module contain all relation with orm tables."""
from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .utils import TModelClass

from .orm import MColumn
from .orm import Relation
from .orm import RelationForeignKey
from .orm import TColumn


__all__ = (
    "RelationForeignKey",
    "Relation",
    "TColumn",
    "TModelClass",
    "MColumn",
)
